/*
 * Part of utils for Z-Way server
 *
 * Copyright Z-Wave.Me
 *
 * Author: Serguei Poltorak <ps@z-wave.me>, 2014
 *
 * License: GNU GPLv3
 *
 * Usage:
 *  /JS/Run/executeFile("z-way-utils/CreateZDDX.js")
 *  /ZAutomation/CreateZDDX/nodeId
 *
 */

// Defines global function to handle HTTP requests

function CreateZDDX (uri, request) {
	function hexByteToStr(n) {
		return ("00" + parseInt(n).toString(16)).slice(-2);
	}

	function hexWordToStr(n) {
		return ("0000" + parseInt(n).toString(16)).slice(-4);
	}

	function tagAttrValue(name, value) {
		return	{
			"name": name,
			"attributes": {
				"value": value
			}
		};
	}

	function tagByte(name, value) {
		return tagAttrValue(name, hexByteToStr(value));
	}

	function tagWord(name, value) {
		return tagAttrValue(name, hexWordToStr(value));
	}

	function tagBool(name, value) {
		return tagAttrValue(name, value ? "true": "false");
	}

	function tagText(name, value) {
		return	{
			"name": name,
			"text": value
		};
	}

	function tagLangs(name, values) {
		var
			lang,
			langChildren = [];

		for (lang in values) {
			langChildren.push({
				"name": "lang",
				"attributes": {
					"xml:lang": lang
				},
				"text": values[lang]
			});
		}

		return {
			"name": name,
			"children": langChildren
		};
	}

	function inNIF(id, nif, afterMark) {
		var
			ccId,
			markFound = false;
		
		id = parseInt(id);
		for (i in nif) {
			if (nif[i] === 0xEF) {
				markFound = true;
			}
			
			if (!(afterMark ^ markFound) && parseInt(nif[i]) === id) {
				return true;
			}
		}
		return false;	
	}

	function tagCC(id, version, supported, secure, nif) {
		return {
			"name": "commandClass",
			"attributes": {
				"id": hexWordToStr(id),
				"version": version,
				"supported": supported || inNIF(id, nif, false),
				"controlled": !supported || inNIF(id, nif, true),
				"secure": secure,
				"inNIF": (supported && inNIF(id, nif, false)) || (!supported && inNIF(id, nif, true))
			}
		};
	}

	var	nodeId = uri.split("/")[1],
		d = zway.devices[nodeId],
		zddx = new ZXmlDocument();
		
	zddx.root = {
		"name": "ZWaveDevice",
		"attributes": {
			"xmlns": "http://www.pepper1.net/zwavedb/xml-schemata/z-wave",
			"schemaVersion": "2"
		},
		"children": [
			{
				"name": "descriptorVersion",
				"text": "1"
			},
			{
				"name": "deviceData",
				"children": [
					tagWord("manufacturerId", d.data.manufacturerId.value),
					tagWord("productType", d.data.manufacturerProductType.value),
					tagWord("productId", d.data.manufacturerProductId.value),
					tagByte("libType", d.data.ZWLib.value),
					tagByte("protoVersion", d.data.ZWProtocolMajor.value),
					tagByte("protoSubVersion", d.data.ZWProtocolMinor.value),
					tagByte("appVersion", d.data.applicationMajor.value),
					tagByte("appSubVersion", d.data.applicationMinor.value),
					tagByte("basicClass", d.data.basicType.value),
					tagByte("genericClass", d.data.genericType.value),
					tagByte("specificClass", d.data.specificType.value),
					tagBool("optional", d.data.optional.value),
					tagBool("listening", d.data.isListening.value),
					tagBool("routing", d.data.isRouting.value),
					tagText("beamSensor", d.data.sensor250.value ? "250" : (d.data.sensor1000.value ? "1000" : "0")),
				]
			},
			{
				"name": "deviceDescription",
				"children": [
					tagLangs("description", {"en": ""}),
                                        tagLangs("wakeupNote", {"en": ""}),
                                        tagLangs("inclusionNote", {"en": ""}),
                                        tagText("productName", ""),
                                        tagText("brandName", ""),
					tagText("productVersion", d.data.applicationMajor.value.toString()  + "." + d.data.applicationMinor.value.toString())
				]
			},
			{
				"name": "commandClasses",
				"children": (function() {
					var
						ccId,
						arr = [],
						ccs = d.instances[0].commandClasses;
					
					for(ccId in ccs) {
						arr.push(tagCC(ccId, ccs[ccId].data.version.value, ccs[ccId].data.supported.value, ccs[ccId].data.security.value, d.data.nodeInfoFrame.value));
					};
					for(n in d.data.nodeInfoFrame.value) {
						var ccId = d.data.nodeInfoFrame.value[n];
						if (!ccs[ccId] && ccId != 0xEF) {
							arr.push(tagCC(ccId, 1, false, false, d.data.nodeInfoFrame.value));
						}
					};
					
					return arr;
				})()
			}
		]
	};
	
	if (d.instances[0].Association) {
		console.logJS(zddx.root.children);
		zddx.root.insertChild({
			"name": "assocGroups",
			"children": (function(data) {
				var
					n,
					Assocs = [];
				
				for (n = 1; n <= data.groups.value; n++) {
					Assocs.push({
						"name": "assocGroup",
						"attributes": {
							"number": n,
							"maxNodes": data[n].max.value
						},
						"children": [
							tagLangs("description", {"en": "Group " + n.toString()})
						]
					});
				}
				
				return Assocs;
			})(d.instances[0].Association.data)
		});
		console.logJS(zddx.root.children);
	}
	
	return {
		"status": 200,
		"body": zddx.toString(),
		"headers": {
			"Content-Type": "application/xml"
		}
	};
};
